/*
 * Copyright 2004-2010 Information & Software Engineering Group (188/1)
 *                     Institute of Software Technology and Interactive Systems
 *                     Vienna University of Technology, Austria
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.ifs.tuwien.ac.at/mir/index.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package at.tuwien.ifs.feature.extraction.audio;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;

import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.JProgressBar;

import at.tuwien.ifs.feature.extraction.audio.data.FeatureExtractionOptions;
import at.tuwien.ifs.feature.extraction.audio.data.RealMatrixExt;
import at.tuwien.ifs.feature.extraction.audio.input.AudioFileSegmentReader;
import at.tuwien.ifs.feature.extraction.audio.util.MyLogger;
import at.tuwien.ifs.feature.extraction.audio.vis.MatrixVisWindow;

/**
 * Copyright Vienna University of Technology
 * 
 * @author Thomas Lidy
 * @version $Id: AudioFileExtractor.java 195 2010-08-17 16:52:18Z frank $
 */
public class AudioFileExtractor {

    public RealMatrixExt[] extractAudioFile(InputStream inStream, FeatureExtractionOptions opt)
            throws UnsupportedAudioFileException, IOException, FeatureExtractorException {
        return extractAudioFile(AudioSystem.getAudioInputStream(inStream), opt, null);

    }

    public RealMatrixExt[] extractAudioFile(File fileIn, FeatureExtractionOptions opt,
            JProgressBar progressBar) throws UnsupportedAudioFileException, IOException,
            FeatureExtractorException {
        MyLogger.info("Extracting \"" + fileIn.toString() + "\"");
        if (progressBar != null) {
            progressBar.setString(fileIn.getName());
        }
        return extractAudioFile(AudioSystem.getAudioInputStream(fileIn), opt, progressBar);
    }

    public RealMatrixExt[] extractAudioFile(String filename, FeatureExtractionOptions opt)
            throws UnsupportedAudioFileException, IOException, FeatureExtractorException {
        File fileIn = new File(filename);
        return extractAudioFile(fileIn, opt);
    }

    public RealMatrixExt[] extractAudioFile(File fileIn, FeatureExtractionOptions opt)
            throws UnsupportedAudioFileException, IOException, FeatureExtractorException {
        return extractAudioFile(fileIn, opt, null);
    }

    public RealMatrixExt[] extractAudioFile(AudioInputStream inStream,
            FeatureExtractionOptions opt, JProgressBar progressBar)
            throws UnsupportedAudioFileException, IOException, FeatureExtractorException {
        @SuppressWarnings("unused")
        int preferredNumSegments = 6;

        AudioFileSegmentReader audioFile;

        // try {
        audioFile = new AudioFileSegmentReader(inStream);
        /*
         * } catch (UnsupportedAudioFileException e) { System.err.println("Unsupported audio file: "
         * + fileIn.getName()); return; } catch (IOException e) { System.err.println("IOException: "
         * + fileIn.getName()); return; }
         */
        long fileLength = audioFile.getNumberOfSamples();
        int sampleRate = audioFile.getSampleRate();
        int sampleBits = audioFile.getSampleSizeInBits();
        int segmentSize;
        int fftWindowSize;

        // segment_size should always be ~6 sec, fft_window_size should always
        // be ~ 23ms
        if (sampleRate <= 11025) {
            segmentSize = 65536 /* 2^16 */;
            fftWindowSize = 256;
        } else if (sampleRate <= 22050) {
            segmentSize = 131072 /* 2^17 */;
            fftWindowSize = 512;
        } else if (sampleRate <= 44100) {
            segmentSize = 262144 /* 2^18 */;
            fftWindowSize = 1024;
        } else
            throw new FeatureExtractorException("Sample rate of " + sampleRate
                    + " Hz not supported!");

        // int bandLimit = BarkScale.getNumberOfBarkBands(sampleRate);

        int numSegmentsInFile = (int) (fileLength / segmentSize);

        if (numSegmentsInFile == 0)
            throw new FeatureExtractorException("Number of segments will be 0! (File length: "
                    + fileLength + "  Segment Size: " + segmentSize + ")");

        int stepWidth = opt.stepWidth;
        int skipSegments = opt.skipLeadInFadeOut;
        /*
         * adjust stepWidth so that a preferred number of segments is read and the rest are skipped
         */
        // int stepWidth = numSegmentsInFile / preferredNumSegments;
        // if (stepWidth == 0) stepWidth = 1;
        if ((skipSegments > 0) | (stepWidth > 1)) {
            if (audioFile.getDurationInSeconds() < 45) {
                stepWidth = 1;
                skipSegments = 0;
                MyLogger.warning("Track is < 45 sec. Ignoring step width > 1 and skip leadin/fadeout options.");
            }
        }

        int numFSets = opt.getNumberOfFeatureSets();
        // int numSegmentsToExtract = numSegmentsInFile / stepWidth;
        int numSegmentsToExtract = ((numSegmentsInFile - (skipSegments * 2) - 1) / stepWidth) + 1;

        MyLogger.extrainfo("FileLength(Frames): " + fileLength + "  Segment Size: " + segmentSize
                + "  # of Segments: " + numSegmentsInFile + "  Step Width: " + stepWidth
                + "  Skip: " + skipSegments + "  Segments to extract: " + numSegmentsToExtract);

        RealMatrixExt[][] featureSets = new RealMatrixExt[numSegmentsToExtract][];

        FeatureExtractor featExtract = new FeatureExtractor(sampleRate, sampleBits);

        int seg; // segment# read in
        int extseg = 0; // segment# extracted
        int framesRead; // #frames read from file in one iteration
        int totalFramesRead = 0; // total #frames read from file

        short[] waveSegment = new short[segmentSize];

        if (progressBar != null) {
            progressBar.setMaximum(numSegmentsToExtract);
            progressBar.setValue(0);
        }
        MyLogger.extrainfo("Segment");

        for (seg = 0; (seg < numSegmentsInFile) && (extseg < numSegmentsToExtract); seg++) {
            try {
                framesRead = audioFile.readMono(waveSegment);
            } catch (IOException e) {
                throw new IOException("Problem reading segment " + seg + ". " + e.getMessage());
            }

            MyLogger.extrainfoNoNewLine(" " + seg);
            // System.out.println("Segment " + seg + " FramesRead: " +
            // framesRead);

            totalFramesRead += framesRead;

            if ((framesRead == -1) || (framesRead < segmentSize)) {
                break; // end of file
            }

            // for (int i=0; i < segmentSize; i++)
            // {
            // System.out.println(waveSegment[i]);
            // }

            // start with x-th segment (according to skipSegments),
            // take only every 2nd, 3rd, 4th, ... segment according to stepWidth
            if (((seg - skipSegments) >= 0) && ((seg - skipSegments) % stepWidth == 0)) {
                MyLogger.extrainfoNoNewLine("*");
                // result[extseg] =
                // featExtract.extractRhythmPatterns(waveSegment, bandLimit,
                // modAmplLimit, fftWindowSize);
                featureSets[extseg] = featExtract.extractFeatureSets(waveSegment, opt,
                        fftWindowSize);
                extseg++;
                if (progressBar != null) {
                    progressBar.setValue(extseg);
                }
            }
        }

        audioFile.closeStream();

        int segmentsRead = extseg;
        // MyLogger.extrainfo(""); // new line
        MyLogger.extrainfo("Segments extracted: " + segmentsRead + " totalFramesRead: "
                + totalFramesRead);

        if (segmentsRead == 0)
            throw new FeatureExtractorException(
                    "Number of segments extracted was 0! (File length: " + fileLength
                            + "  Segment Size: " + segmentSize + "  Step Width: " + stepWidth);
        else if (segmentsRead != numSegmentsToExtract) {
            // FIXME(?): throw new FeatureExtractorException
            MyLogger.warning("Number of segments extracted (" + segmentsRead
                    + ") differs from expected number (" + numSegmentsToExtract + ")! "
                    + "(File length: " + fileLength + "  Segment Size: " + segmentSize
                    + "  Step Width: " + stepWidth);
        }

        /* calculate Median */

        RealMatrixExt[] medianFeatureSets = new RealMatrixExt[numFSets];

        for (int fi = 0; fi < numFSets; fi++) {
            // "transpose"
            RealMatrixExt[] segmFeatureSets = new RealMatrixExt[segmentsRead];
            for (int s = 0; s < segmentsRead; s++) {
                segmFeatureSets[s] = featureSets[s][fi];
            }

            medianFeatureSets[fi] = RealMatrixExt.median(segmFeatureSets);

            if (opt.showFeatureVis) {
                new MatrixVisWindow(opt.visParentFrame, medianFeatureSets[fi].getData(),
                        "MEDIAN feature set " + medianFeatureSets[fi].getType());
            }
        }

        return medianFeatureSets;
    }

}
