/*
 * Copyright 2004-2010 Information & Software Engineering Group (188/1)
 *                     Institute of Software Technology and Interactive Systems
 *                     Vienna University of Technology, Austria
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.ifs.tuwien.ac.at/mir/index.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package at.tuwien.ifs.feature.extraction.audio;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Collections;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;

import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.JProgressBar;

import at.tuwien.ifs.feature.extraction.audio.MainGUI.Status;
import at.tuwien.ifs.feature.extraction.audio.data.FeatureExtractionOptions;
import at.tuwien.ifs.feature.extraction.audio.data.FeatureVectorData;
import at.tuwien.ifs.feature.extraction.audio.data.RealMatrixExt;
import at.tuwien.ifs.feature.extraction.audio.output.AbstractDataWriter;
import at.tuwien.ifs.feature.extraction.audio.output.ClassAssignmentResolver;
import at.tuwien.ifs.feature.extraction.audio.util.MyLogger;
import at.tuwien.ifs.somtoolbox.data.InputDataWriter;
import at.tuwien.ifs.somtoolbox.input.SOMLibFileFormatException;
import at.tuwien.ifs.somtoolbox.util.DateUtils;

public class ExtractionDispatcher {

    private class ExtractionWorker extends Thread {
        private final ExtractionDispatcher dispatcher;

        private final FeatureExtractionOptions options;

        private final MainGUI gui;

        @SuppressWarnings("unused")
        public ExtractionWorker(ExtractionDispatcher d, FeatureExtractionOptions options) {
            this(d, options, null);
        }

        public ExtractionWorker(ExtractionDispatcher d, FeatureExtractionOptions options,
                MainGUI gui) {
            this.dispatcher = d;
            this.options = options;
            this.gui = gui;

            this.setName(ExtractionWorker.class.getSimpleName());
        }

        @Override
        public void run() {
            JProgressBar progress = null;
            if (gui != null) {
                progress = gui.createThreadProgressBar();
            }
            AudioFileExtractor afx = new AudioFileExtractor();
            RealMatrixExt[] featureSets = null;

            for (FeatureVectorData data = dispatcher.getExtractionJob(); data != null; data = dispatcher.getExtractionJob()) {
                boolean error = false;
                String relFilename = data.getFileLabel();
                File curFile = new File(MultipleFileExtractor.getAbsoluteFilename(relFilename,
                        options.inputBaseDir));
                long fileSize = curFile.length();

                try {
                    featureSets = afx.extractAudioFile(curFile, options, progress);
                } catch (UnsupportedAudioFileException e) {
                    MyLogger.error("Unsupported audio file: " + curFile + " - " + e.getMessage());
                    error = true;
                } catch (IOException e) {
                    MyLogger.error("IOException: " + curFile + " - " + e.getMessage());
                    error = true;
                } catch (FeatureExtractorException e) {
                    MyLogger.error("Exception during feature extraction: " + curFile + " - "
                            + e.getMessage());
                    error = true;
                } catch (Exception e) {
                    MyLogger.error("Unknown Exception during feature extraction: " + curFile
                            + " - " + e.getMessage());
                    error = true;
                }
                if (error) {
                    MyLogger.error("Extracting \"" + curFile.toString() + "\" failed.");
                }
                MyLogger.info("");
                MyLogger.flush();
                dispatcher.putExtractionResult(data, featureSets, fileSize, error);
            }

            // Exit and clean up.
            if (gui != null) {
                gui.removeThreadProgressBar(progress);
            }
            dispatcher.workers.remove(this);
        }
    }

    private List<FeatureVectorData> todoList;

    private FeatureExtractionOptions options;

    private final MainGUI gui;

    private List<ExtractionWorker> workers;

    private int itemsInProgress = 0;

    private boolean isRunning = false;

    private int desiredThreadCount = 0;

    private static final int MINIMAL_THREADS = 1;

    private AbstractDataWriter[][] outputFiles;

    int numErrors = 0;

    long startTime = 0;

    StringBuilder errorLog = new StringBuilder();

    private final PrintStream progress;

    private int fileCount, filesDone;

    private long byteCount, bytesDone;

    private ClassAssignmentResolver classAssignmentResolver = new ClassAssignmentResolver();

    private boolean wasCancelled;

    public ExtractionDispatcher(List<FeatureVectorData> todoList, FeatureExtractionOptions options,
            AbstractDataWriter[][] outputFiles, boolean silent) {
        this(todoList, -1, options, outputFiles, silent, null);
    }

    public ExtractionDispatcher(List<FeatureVectorData> todoList, long byteCount,
            FeatureExtractionOptions options, AbstractDataWriter[][] outputFiles, boolean silent) {
        this(todoList, byteCount, options, outputFiles, silent, null);
    }

    public ExtractionDispatcher(List<FeatureVectorData> todoList, long byteCount,
            FeatureExtractionOptions options, AbstractDataWriter[][] outputFiles, boolean silent,
            MainGUI gui) {
        this.gui = gui;
        this.todoList = todoList;
        this.options = options;
        this.outputFiles = outputFiles;

        progress = silent ? null : System.out;
        this.fileCount = todoList.size();
        this.filesDone = 0;
        this.bytesDone = 0;
        this.byteCount = byteCount;

        if (options.progressPanel != null) {
            options.progressPanel.setTotalFilesProgressMaximum(fileCount);
        }

        workers = Collections.synchronizedList(new LinkedList<ExtractionWorker>());
        this.wasCancelled = false;
    }

    /**
     * Start additional threads
     * 
     * @param threads
     */
    public synchronized void startThreads(int threads) {
        if (!isRunning) {
            startTime = System.currentTimeMillis();
        }

        isRunning = true;
        desiredThreadCount += threads;
        MyLogger.info(String.format("Starting %d additional threads", threads));
        MyLogger.flush();
        for (int i = 0; i < threads; i++) {
            ExtractionWorker t = new ExtractionWorker(this, options, gui);
            workers.add(t);
            t.start();
        }
    }

    public synchronized void stopThreads(int threads) {
        // At least MINIMAL_THREADS must survive!
        MyLogger.info(String.format("Trying to stop %d threads", threads));
        MyLogger.flush();
        desiredThreadCount = Math.max(desiredThreadCount - threads, MINIMAL_THREADS);
    }

    public synchronized void setDesiredThreadCount(int threads) {
        if (threads > desiredThreadCount) {
            startThreads(threads - desiredThreadCount);
        } else if (threads < desiredThreadCount) {
            stopThreads(desiredThreadCount - threads);
        }
    }

    /**
     * Start the extraction. This method blocks until the extraction is finished!
     * 
     * @param threads define how many parallel extraction jobs will be started.
     */
    public boolean extract(int threads) {
        MyLogger.info(String.format("Starting extraction at %tF %<tT with %d Threads", new Date(),
                threads));
        if (gui != null) {
            gui.getOverallProgress().setMaximum(fileCount);
        }
        startThreads(threads);

        // Block until work is done...
        while (getItemsInProgress() > 0 || todoList.size() > 0) {
            try {
                Thread.sleep(1000);

                synchronized (workers) {
                    for (int i = workers.size(); i < desiredThreadCount; i++) {
                        ExtractionWorker t = new ExtractionWorker(this, options, gui);
                        workers.add(t);
                        t.start();
                    }
                }

            } catch (InterruptedException e) {
                // NOP
            }
        }
        
        if (wasCancelled) {
            return false;
        }

        long stopTime = System.currentTimeMillis();
        if (gui != null) {
            gui.getOverallProgress().setValue(gui.getOverallProgress().getMaximum());
        }

        // resolve class assignment
        classAssignmentResolver.resolveClassAssignment();

        if (options.writeSOMLib) {
            try {
                InputDataWriter.writeAsSOMLib(classAssignmentResolver.getClassAssignments(),
                        classAssignmentResolver.getClassNames(), options.outputBaseFilename
                                + ".cls");
            } catch (SOMLibFileFormatException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            } catch (IOException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
        }

        if (outputFiles != null) {
            try {
                // tell all writers to finish
                for (AbstractDataWriter[] outputFile : outputFiles) {
                    for (AbstractDataWriter element : outputFile) {
                        // set class assignment resolver
                        element.setClassAssignmentResolver(classAssignmentResolver);

                        element.close();
                    }
                }
            } catch (IOException e) {
                MyLogger.error("IOException while trying to close vector file: " + e.getMessage());
            }
        }
        // printProgressBar();
        if (progress != null) {
            progress.println();
            progress.printf("Extracted %d of %d Songs (%d Errors) in %s%n", filesDone, fileCount,
                    numErrors, DateUtils.shortFormatDuration(stopTime - startTime));
            if (numErrors > 0) {
                progress.printf("Theses %d files produced errors:%n%s", numErrors, errorLog);
            }
            progress.println();
        }
        if (gui != null) {
            gui.setStatus(Status.DONE);
        }
        return true;
    }

    public synchronized FeatureVectorData getExtractionJob() {
        if (todoList.size() == 0) {
            return null;
        }

        if (desiredThreadCount < workers.size()) {
            return null;
        }

        adjustItemsInProgress(1);
        printProgressBar();
        return todoList.remove(0);
    }

    private void printProgressBar() {
        if (progress == null) {
            return;
        }
        long now = System.currentTimeMillis();
        long todo = 0;
        double percent = 0f;
        long eta = now;
        if (filesDone < fileCount) {
            if (byteCount < 0) {
                if (filesDone > 0) {
                    percent = (double) filesDone / (double) fileCount;
                    eta = startTime + Math.round((now - startTime) / percent);
                    todo = eta - now;
                }
            } else {
                if (bytesDone > 0) {
                    percent = (double) bytesDone / (double) byteCount;
                    eta = startTime + Math.round((now - startTime) / percent);
                    todo = eta - now;
                }
            }

            if (options.progressPanel != null) {
                options.progressPanel.setTotalFilesProgress(filesDone);
            }
        } else {
            percent = 1f;
        }

        String progressString = String.format("\r(%.1f%%) %d/%d (%d) ETA: %tF %<tT (%s) ",
                percent * 100f, filesDone, fileCount, getItemsInProgress(), new Date(eta),
                DateUtils.shortFormatDuration(todo));
        progress.print(progressString);
        if (gui != null) {
            gui.getOverallProgress().setValue(filesDone);
            gui.getOverallProgress().setString(String.format("ETA: %tF %<tT", new Date(eta)));
        }

        if (MultipleFileExtractor.verboose) {
            try {
                String fname = options.outputBaseFilename + ".timelog";
                PrintStream ps;
                ps = new PrintStream(new FileOutputStream(fname, true));
                ps.println(progressString);
                ps.close();
            } catch (FileNotFoundException e) {
            }
            try {
                String fname = options.outputBaseFilename + ".rawtimelog";
                PrintStream ps;
                ps = new PrintStream(new FileOutputStream(fname, true));
                ps.printf("%d %.5f %d %d %d %d %d %d%n", now, percent, filesDone, fileCount,
                        bytesDone, byteCount, eta, todo);
                ps.close();
            } catch (FileNotFoundException e) {
            }

        }
    }

    private int getItemsInProgress() {
        return adjustItemsInProgress(0);
    }

    private synchronized int adjustItemsInProgress(int change) {
        itemsInProgress += change;

        return itemsInProgress;
    }

    public synchronized void putExtractionResult(FeatureVectorData item, RealMatrixExt[] features,
            long fileSize, boolean error) {
        adjustItemsInProgress(-1);
        filesDone++;
        bytesDone += fileSize;
        printProgressBar();
        // Write...
        if (!error && outputFiles != null) {
            try {
                for (int feat = 0; feat < features.length; feat++) {
                    item.setMatrix(features[feat]);
                    for (int file = 0; file < outputFiles[feat].length; file++) {
                        // outputFiles[feat][file].writeMatrix(featureSets[feat],relFilename);
                        outputFiles[feat][file].writeVector(item);
                    }
                }

            } catch (IOException e) {
                MyLogger.severe("IOException while writing to vector file: " + e.getMessage());
                error = true;
            }
            classAssignmentResolver.addLabel(item.getFileLabel());
        } else {
            errorLog.append(String.format("  %s%n", item.getFileLabel()));
        }

        if (error) {
            numErrors++;
        }

        // This is currently done via blocking start-method...
        // if ((adjustItemsInProgress(0) == 0) && (todoList.size() == 0)) {
        // }

    }

    /**
     * 
     */
    public void cancelExtraction() {
        wasCancelled = true;
        todoList.clear();
    }
}
