/*
 * Copyright 2004-2010 Information & Software Engineering Group (188/1)
 *                     Institute of Software Technology and Interactive Systems
 *                     Vienna University of Technology, Austria
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.ifs.tuwien.ac.at/mir/index.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package at.tuwien.ifs.feature.extraction.audio;

import java.util.Arrays;

import org.apache.commons.math.linear.RealMatrix;
import org.apache.commons.math.stat.descriptive.moment.Kurtosis;
import org.apache.commons.math.stat.descriptive.moment.Mean;
import org.apache.commons.math.stat.descriptive.moment.Skewness;
import org.apache.commons.math.stat.descriptive.moment.Variance;
import org.imirsel.m2k.util.Window;

import at.tuwien.ifs.feature.extraction.audio.data.FeatureExtractionOptions;
import at.tuwien.ifs.feature.extraction.audio.data.RealMatrixExt;
import at.tuwien.ifs.feature.extraction.audio.spectrum.BarkScale;
import at.tuwien.ifs.feature.extraction.audio.spectrum.FFT;
import at.tuwien.ifs.feature.extraction.audio.spectrum.Spectrogram;
import at.tuwien.ifs.feature.extraction.audio.vis.MatrixVisWindow;

/**
 * Copyright Vienna University of Technology
 * 
 * @author Thomas Lidy
 * @version $Id: FeatureExtractor.java 167 2010-06-17 10:12:01Z frank $
 */
public class FeatureExtractor {

    private boolean DEBUG = false;

    private int sampleRate;

    private int sampleSizeInBits;

    /**
     * @param sampleRate the sample rate
     * @param sampleSizeInBits the sample size in bits
     */
    public FeatureExtractor(int sampleRate, int sampleSizeInBits) {
        this.sampleRate = sampleRate;
        this.sampleSizeInBits = sampleSizeInBits;
    }

    /**
     * @param wavesegment audio data
     * @param opt Feature Extraction Options fftWindowSize size of window for FFT calculation will
     *            be determined from sampleRate
     * @return RealMatrixExt array containing one RealMatrixExt for each feature set extracted
     */
    public RealMatrixExt[] extractFeatureSets(short[] wavesegment, FeatureExtractionOptions opt) {
        int fft_window_size = 1024;
        if (sampleRate <= 11025) {
            fft_window_size = 256;
        } else if (sampleRate <= 22050) {
            fft_window_size = 512;
        }
        // else { fft_window_size = 1024; }

        return extractFeatureSets(wavesegment, opt, fft_window_size);
    }

    /**
     * @param waveSegment audio data
     * @param opt Feature Extraction Options
     * @param fftWindowSize size of window for FFT calculation
     * @return RealMatrixExt array containing one RealMatrixExt for each feature set extracted
     */
    public RealMatrixExt[] extractFeatureSets(short[] waveSegment, FeatureExtractionOptions opt,
            int fftWindowSize) {
        int segmentSize = waveSegment.length;
        int fi = 0; // feature set index
        RealMatrixExt[] featureSets = new RealMatrixExt[opt.getNumberOfFeatureSets()];

        /* convert to double precision and */
        /* scale wave data to adjust according to hearing threshold */

        @SuppressWarnings("unused")
        double divisor;

        if (sampleSizeInBits == 8) {
            divisor = 256;
        } else {
            divisor = 32768;
        }

        double[] wav = new double[segmentSize];

        for (int i = 0; i < segmentSize; i++) {
            // TODO: investigate factor
            wav[i] = waveSegment[i]; // / divisor * 50; //0.0875 * 32768; //
            // (0.0875 * 2^15)
            // System.out.println(wav[i]);
        }

        /*
         * compute spectrogram for each wave segment (with 50 % overlap in FFT computation)
         */

        double[][] spec = Spectrogram.computeSpectrogram(wav, fftWindowSize, fftWindowSize / 2,
                Window.HAMMING);

        // if (spec == null)
        // {
        // //TODO throw Exception
        // System.err.println("Spectrogram returned null!");
        // return;
        // }

        if (opt.showIntermediateVis) {
            new MatrixVisWindow(opt.visParentFrame, spec, "Spectrogram");
        }

        int nFrames = spec.length;

        if (DEBUG) {
            System.out.println("spec: " + spec.length + "x" + spec[0].length);

            for (int w = 0; w < spec.length; w++) {
                System.out.print("[t" + w + "] ");
                for (int b = 0; b < spec[0].length; b++) {
                    System.out.print(b + ": " + spec[w][b] + "  ");
                }
                System.out.println();
            }
        }

        /* group to bark bands */

        BarkScale barkScale = new BarkScale(sampleRate, fftWindowSize);
        int nBands = barkScale.getNumberOfBarkBands();

        double[][] specBark = new double[nFrames][nBands];

        for (int t = 0; t < nFrames; t++) {
            specBark[t] = barkScale.apply(spec[t]);
        }

        if (opt.showIntermediateVis) {
            new MatrixVisWindow(opt.visParentFrame, specBark, "Bark Scale");
        }

        if (DEBUG) {
            System.out.println("bark: " + specBark.length + "x" + specBark[0].length);

            for (int t = 0; t < nFrames; t++) {
                System.out.print("[t" + t + "] ");
                for (int b = 0; b < specBark[0].length; b++) {
                    System.out.print(b + ": " + specBark[t][b] + "  ");
                }
                System.out.println();
            }
        }

        /* decibel */
        for (int t = 0; t < nFrames; t++) {
            for (int b = 0; b < nBands; b++) {
                if (specBark[t][b] < 1) {
                    specBark[t][b] = 0; // in Matlab code this is 1
                } else {
                    specBark[t][b] = 10.0 * Math.log10(specBark[t][b]); // in
                    // Simon's
                    // M2K
                    // code
                    // this
                    // was
                    // log
                }
            }
        }

        if (opt.showIntermediateVis) {
            new MatrixVisWindow(opt.visParentFrame, specBark, "Decibel Scale");
        }

        if (DEBUG) {
            System.out.println("decibel: ");

            for (int t = 0; t < nFrames; t++) {
                System.out.print("[t" + t + "] ");
                for (int b = 0; b < specBark[0].length; b++) {
                    System.out.print(b + ": " + specBark[t][b] + "  ");
                }
                System.out.println();
            }
        }

        /* Phon */
        // TODO

        /* Sone */

        for (int t = 0; t < nFrames; t++) {
            for (int b = 0; b < nBands; b++) {
                if (specBark[t][b] >= 40) {
                    specBark[t][b] = Math.pow(2, ((specBark[t][b] - 40) / 10));
                } else {
                    specBark[t][b] = Math.pow((specBark[t][b] / 40), 2.642);
                }
            }
        }

        if (opt.showIntermediateVis) {
            new MatrixVisWindow(opt.visParentFrame, specBark, "Sone Scale");
        }

        /* transpose matrix (for efficient row processing) */

        RealMatrix specBarkT = new RealMatrixExt(specBark).transpose();

        // if (opt.showIntermediateVis) new MatrixVisWindow(opt.visParentFrame,
        // specBarkT, "Sone Scale (transposed)");

        /* compute SSD features */

        if (opt.extractSSD) {
            featureSets[fi++] = computeStatisticalSpectrumDescriptor(specBarkT, opt.bandLimit);
        }

        if (!(opt.extractRP || opt.extractRH))
            return featureSets;

        /* FFT */

        double[][] specModAmp = new double[nBands][];

        FFT FFTreal = new FFT();

        for (int b = 0; b < nBands; b++) {
            // compute FFT magnitude of each band
            specModAmp[b] = FFTreal.computeMagnitude(specBarkT.getRow(b));
        }

        if (opt.showIntermediateVis) {
            new MatrixVisWindow(opt.visParentFrame, specModAmp, "Modulation Amp (transposed)");
        }

        /*
         * if (DEBUG) { for (int b = 0; b < nBands; b++) { System.out.print("[b" + b + "] "); for
         * (int t = 0; t < nFrames; t++) { System.out.print(specBarkT[b][t] + " "); }
         * System.out.println(); } }
         */

        /* Fluctuation Strength Curve */
        // TODO: possible optimizations:
        // - precalc fluctWeight in an array
        // - calculate only < bandLimit <= modAmplLimit

        // resolution of modulation frequency axis (0.17 Hz)
        double modulationFreqResolution = 1 / ((double) segmentSize / (double) sampleRate);

        for (int b = 0; b < nBands; b++) {
            specModAmp[b][0] = 0; // omit DC component (would cause deviations
            // while blurring)
            for (int t = 1; t <= nFrames; t++) // skip DC component (avoid
            // div/0)
            {
                double modFreq = modulationFreqResolution * t;
                double fluctWeight = 1 / (modFreq / 4 + 4 / modFreq);
                specModAmp[b][t] = specModAmp[b][t] * fluctWeight;
            }
        }

        if (opt.showIntermediateVis) {
            new MatrixVisWindow(opt.visParentFrame, specModAmp,
                    "Fluctuation Weighting (transposed)");
        }

        /* Blurring */

        // TODO

        /* crop (or fill) matrix for result */

        int bandLimit = opt.bandLimit;
        double[][] result = new double[opt.bandLimit][opt.modAmplLimit];

        if (nBands < opt.bandLimit) // requested number of bands is > than
        // actual # of bands
        { // fill remainder with 0
            for (int b = nBands; b < opt.bandLimit; b++) {
                Arrays.fill(result[b], 0);
            }
            bandLimit = nBands;
        }

        for (int b = 0; b < bandLimit; b++) {
            for (int t = 1; t <= opt.modAmplLimit; t++) // skip DC component:
            // start with 1
            {
                result[b][t - 1] = specModAmp[b][t];
            }
        }

        RealMatrixExt rhythmPattern = new RealMatrixExt(result);
        rhythmPattern.setType(RealMatrixExt.TYPE_RP);

        if (opt.extractRP) {
            featureSets[fi++] = rhythmPattern;
            if (opt.showIntermediateVis) {
                new MatrixVisWindow(opt.visParentFrame, result, "Rhythm Pattern");
            }
        }

        /* compute RH features */

        if (opt.extractRH) {
            featureSets[fi++] = computeRhythmHistogram(rhythmPattern);
            // if (opt.showIntermediateVis) new
            // LineVisWindow(/*opt.visParentFrame,*/
            // featureSets[fi-1].getRow(0), "Rhythm Pattern");
        }

        return featureSets;
    }

    private RealMatrixExt computeRhythmHistogram(RealMatrixExt rhythmPattern) {
        /* create row vector -> matrix with 1 row only */
        double[][] vector = new double[1][];
        vector[0] = rhythmPattern.sumColumns();
        RealMatrixExt result = new RealMatrixExt(vector);
        result.setType(RealMatrixExt.TYPE_RH);
        return result;
    }

    private static int N_STATISTICAL_DESCRIPTORS = 7;

    private RealMatrixExt computeStatisticalSpectrumDescriptor(RealMatrix sonogram, int bandLimit) {
        double[][] matrixSSD = new double[bandLimit][N_STATISTICAL_DESCRIPTORS];
        int nRows = sonogram.getRowDimension();

        if (nRows < bandLimit) // requested number of bands is > than actual #
        // of bands -> fill remainder with 0
        {
            for (int r = nRows; r < bandLimit; r++) {
                Arrays.fill(matrixSSD[r], 0);
            }
            bandLimit = nRows;
        }

        for (int r = 0; r < bandLimit; r++) {
            matrixSSD[r] = getStatisticalSpectrumDescriptor(sonogram.getRow(r));
        }

        RealMatrixExt result = new RealMatrixExt(matrixSSD);
        result.setType(RealMatrixExt.TYPE_SSD);
        return result;
    }

    /*
     * DEACTIVATED because of destructive data change occurig in getStatisticalSpectrumDescriptor
     * private RealMatrixExt computeStatisticalSpectrumDescriptor(double[][] sonogram, int
     * bandLimit) { double[][] matrixSSD = new double[bandLimit][N_STATISTICAL_DESCRIPTORS]; int
     * nRows = sonogram.length; if (nRows < bandLimit) //requested number of bands is > than actual
     * # of bands -> fill remainder with 0 { for (int r = nRows; r < bandLimit; r++) {
     * Arrays.fill(matrixSSD[r],0); } bandLimit = nRows; } for (int r = 0; r < bandLimit; r++) {
     * matrixSSD[r] = getStatisticalSpectrumDescriptor(sonogram[r]); } RealMatrixExt result = new
     * RealMatrixExt(matrixSSD); result.setType(RealMatrixExt.TYPE_SSD); return result; }
     */

    private double[] getStatisticalSpectrumDescriptor(double[] dataRow) {
        int N = dataRow.length;
        double ssd[] = new double[N_STATISTICAL_DESCRIPTORS];

        ssd[0] = new Mean().evaluate(dataRow);
        ssd[1] = new Variance().evaluate(dataRow);
        ssd[2] = new Skewness().evaluate(dataRow);
        ssd[3] = new Kurtosis().evaluate(dataRow);

        // NOTE: be careful, sort changes data!
        // (as sonogram.getRow() above copies the data anyway, in this case
        // there is no problem)
        // (otherwise, use dataRow.clone(); )
        Arrays.sort(dataRow);

        // median
        if (N % 2 == 0) {
            ssd[4] = (dataRow[(N / 2) - 1] + dataRow[(N / 2)]) / 2;
        } else {
            ssd[4] = dataRow[(N - 1) / 2];
        }

        ssd[5] = dataRow[0]; // min value
        ssd[6] = dataRow[N - 1]; // max value

        return ssd;
    }

}
