/*
 * Copyright 2004-2010 Information & Software Engineering Group (188/1)
 *                     Institute of Software Technology and Interactive Systems
 *                     Vienna University of Technology, Austria
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.ifs.tuwien.ac.at/mir/index.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package at.tuwien.ifs.feature.extraction.audio;

import java.awt.GraphicsEnvironment;
import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.text.DateFormat;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.TimeZone;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPException;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Switch;
import com.martiansoftware.jsap.stringparsers.FileStringParser;

import at.tuwien.ifs.feature.extraction.audio.MainGUI.Status;
import at.tuwien.ifs.feature.extraction.audio.data.FeatureExtractionOptions;
import at.tuwien.ifs.feature.extraction.audio.data.FeatureVectorData;
import at.tuwien.ifs.feature.extraction.audio.input.AbstractFileListReader;
import at.tuwien.ifs.feature.extraction.audio.input.OSIRISFileListInputReader;
import at.tuwien.ifs.feature.extraction.audio.input.SimpleFileListReader;
import at.tuwien.ifs.feature.extraction.audio.output.ARFFFormatWriter;
import at.tuwien.ifs.feature.extraction.audio.output.AbstractDataWriter;
import at.tuwien.ifs.feature.extraction.audio.output.OSIRISFormatWriter;
import at.tuwien.ifs.feature.extraction.audio.output.SOMLibFormatWriter;
import at.tuwien.ifs.feature.extraction.audio.output.XMLFormatWriter;
import at.tuwien.ifs.feature.extraction.audio.util.FileLogger;
import at.tuwien.ifs.feature.extraction.audio.util.LogHandler.LogGrade;
import at.tuwien.ifs.feature.extraction.audio.util.MyLogger;
import at.tuwien.ifs.feature.extraction.audio.util.PrintStreamLogger;

/**
 * Copyright Vienna University of Technology
 * 
 * @author Thomas Lidy
 * @author Jakob Frank
 * @version $Id: MultipleFileExtractor.java 216 2011-05-07 11:04:26Z mayer $
 */
public class MultipleFileExtractor {
    public static final String VERSION_STRING = "0.7";

    private static final String[] DEFAULT_AUDIO_EXTENSIONS = { "wav", "au", "mp3" };

    static boolean verboose = false;

    public MultipleFileExtractor(int extractionThreads) {
        parallelThreads = extractionThreads;
    }

    public MultipleFileExtractor() {
        this(1);
    }

    public static void main(String[] args) {
        if (args.length == 0) {
            MainGUI.main(args);
            return;
        }

        try {
            // FIXME: use OptionFactory for JSAP
            PrintStreamLogger defaultLogger = new PrintStreamLogger(System.out);
            MyLogger.registerLogHandler(defaultLogger);

            JSAP jsap = new JSAP();

            try {
                setCMDArgs(jsap);
            } catch (JSAPException e) {
                MyLogger.error(e.getMessage());
                MyLogger.flush();
                return;
            }

            String usageInfo = "USAGE: java " + MultipleFileExtractor.class.getName() + " "
                    + jsap.getUsage();

            JSAPResult param = jsap.parse(args);

            if (param.getBoolean("help")) {
                MyLogger.info(usageInfo);
                MyLogger.info(jsap.getHelp().replaceAll("[\\[\\]()]", "").replaceAll("\\|", ", "));
                MyLogger.flush();
                return;
            }

            if (param.getBoolean("version")) {
                MyLogger.info("AudioFeatureExtraction " + VERSION_STRING);
                MyLogger.info("Developed by the IR Team at IFS (Vienna University of Technology)");
                return;
            }

            if (!param.success()) {
                for (Iterator<?> errs = param.getErrorMessageIterator(); errs.hasNext();) {
                    MyLogger.severe(errs.next().toString());
                }
                MyLogger.info("");
                MyLogger.info(usageInfo);
                MyLogger.flush();
                System.exit(1);
            }

            /* initializing options with default parameters */
            FeatureExtractionOptions opt = new FeatureExtractionOptions();

            if (param.getString("outputBaseFilename") == null) {
                MyLogger.error("No outputBaseFilename provided!");
                MyLogger.info("");
                MyLogger.info(usageInfo);
                MyLogger.flush();
                System.exit(1);
            }

            File baseDir = param.getFile("outputBaseDir");
            if (baseDir == null) {
                baseDir = new File(System.getProperty("user.dir"));
            }
            if (!baseDir.exists()) {
                baseDir.mkdirs();
            }
            opt.outputBaseFilename = new File(baseDir, param.getString("outputBaseFilename")).getAbsolutePath();

            /* set/alter user-defined parameters */
            opt.graphicalEnvironment = true;
            // opt.showFeatureVis = true;
            // opt.showIntermediateVis = true;
            // opt.extractRP = false;
            opt.extractSSD = true;
            // opt.extractRH = false;

            if (param.getBoolean("somlib")) {
                opt.writeSOMLib = true;
                opt.enableSOMLibGZip = false;
            }
            if (param.getBoolean("zsomlib")) {
                opt.writeSOMLib = true;
                opt.enableSOMLibGZip = true;
            }

            if (param.getBoolean("xml")) {
                opt.writeXML = true;
            }
            if (param.getBoolean("osiris")) {
                opt.writeOSIRIS = true;
            }
            if (param.getBoolean("arff")) {
                opt.writeWekaARFF = true;
                opt.enableArffGZip = false;
            }
            if (param.getBoolean("zarff")) {
                opt.writeWekaARFF = true;
                opt.enableArffGZip = true;
            }

            if (param.getBoolean("gz")) {
                opt.enableSOMLibGZip = true;
                opt.enableArffGZip = true;
            }

            if (param.getBoolean("log")) {
                try {
                    MyLogger.registerLogHandler(new FileLogger(opt.outputBaseFilename + ".log"));
                } catch (IOException e) {
                    MyLogger.error("Could not open log file: " + e.getMessage());
                }
            }

            if (param.getBoolean("quiet")) {
                MyLogger.removeLogHandler(defaultLogger);
            }

            if (param.getBoolean("verboose")) {
                MyLogger.setGlobalLogGrade(LogGrade.EXTRA_INFO);
                verboose = true;
            }

            if (param.getBoolean("nogui") || GraphicsEnvironment.isHeadless()) {
                opt.disableGraphics();
            }

            MainGUI mainFrame = null;
            boolean closeGUI = false;

            /* start feature extraction */
            MultipleFileExtractor featureExtractor = new MultipleFileExtractor(
                    param.getInt("threads"));

            if (opt.graphicalEnvironment) {
                mainFrame = new MainGUI(featureExtractor);
                opt.visParentFrame = mainFrame;
                mainFrame.setFeatureExtractionOptions(opt);
                mainFrame.setThreadNumber(featureExtractor.parallelThreads);
                mainFrame.setVisible(true);
            }

            if (param.getString("inputFileList") != null) {
                opt.inputBaseDir = param.getFile("inputDirectory").getAbsolutePath();
                mainFrame.setFeatureExtractionOptions(opt);
                File fileListfile = param.getFile("inputFileList");

                AbstractFileListReader fileListReader;
                try {
                    if (param.getBoolean("inosiris")) {
                        fileListReader = new OSIRISFileListInputReader(fileListfile);
                    } else {
                        fileListReader = new SimpleFileListReader(fileListfile);
                    }
                } catch (IOException e) {
                    MyLogger.severe("Problem reading inputFileList: " + e.getMessage());
                    return;
                } finally {
                    MyLogger.flush();
                }
                MyLogger.flush();
                featureExtractor.extractListOfFiles(fileListReader.getMetaDataList(), opt,
                        mainFrame);
            } else if (param.getString("inputDirectory") != null) {
                MyLogger.flush();
                featureExtractor.extractListOfFiles(
                        createExtractionList(param.getFile("inputDirectory"), opt), opt, mainFrame);
            } else {
                MyLogger.error("No input provided! Use either the inputDirectory or inputFileList parameter!");
                MyLogger.info("");
                MyLogger.info(usageInfo);
                closeGUI = true;
            }

            /* close window and exit */
            if (closeGUI && mainFrame != null) {
                mainFrame.dispose();
            }

        } finally {
            MyLogger.flushAll();
        }
    }

    /**
     * @param jsap
     * @throws JSAPException
     */
    private static void setCMDArgs(JSAP jsap) throws JSAPException {
        // FlaggedOption(String id, StringParser stringParser, String
        // defaultValue, boolean required, char shortFlag, String longFlag)

        // Some defaults
        jsap.registerParameter(new Switch("help", 'h', "help", "show this help"));
        jsap.registerParameter(new Switch("version", 'V', "version", "print version and exit"));
        jsap.registerParameter(new Switch("verboose", 'v', "verboose", "print very much!"));
        jsap.registerParameter(new Switch("quiet", 'q', "quiet",
                "do not write anything to the console"));

        // Basic input
        jsap.registerParameter(new FlaggedOption(
                "inputDirectory",
                FileStringParser.getParser().setMustBeDirectory(true).setMustExist(true),
                null,
                false,
                'd',
                JSAP.NO_LONGFLAG,
                "directory which will be recursed for audio files to extract OR single audio filename OR base directory for files contained in a file list"));
        jsap.registerParameter(new FlaggedOption("inputFileList",
                FileStringParser.getParser().setMustBeFile(true).setMustExist(true), null, false,
                'f', JSAP.NO_LONGFLAG,
                "ASCII file containing filenames of input audio files line by line"));
        jsap.registerParameter(new Switch("inosiris", JSAP.NO_SHORTFLAG, "inosiris",
                "inputFileList is provided in OSIRIS format"));

        // Output setings
        jsap.registerParameter(new FlaggedOption("outputBaseDir",
                FileStringParser.getParser().setMustBeDirectory(true), null, false, 'd',
                JSAP.NO_LONGFLAG, "Directory to save output files."));
        jsap.registerParameter(new FlaggedOption("outputBaseFilename", JSAP.STRING_PARSER, null,
                true, 'o', JSAP.NO_LONGFLAG, "filename stub for output file(s) to be written"));
        jsap.registerParameter(new Switch("log", JSAP.NO_SHORTFLAG, "log",
                "log all output to a logfile, named outputBaseFilename.log"));

        // Features
        jsap.registerParameter(new Switch("somlib", 's', "somlib", "write to SOMLib vector format"));
        jsap.registerParameter(new Switch("zsomlib", 'S', "zsomlib",
                "write to SOMLib vector format (gzipped)"));
        jsap.registerParameter(new Switch("arff", 'a', "arff", "write to WEKA ARFF format"));
        jsap.registerParameter(new Switch("zarff", 'A', "zarff",
                "write to WEKA ARFF format (gzipped)"));
        jsap.registerParameter(new Switch("xml", 'x', "xml", "write to XML format"));
        jsap.registerParameter(new Switch("osiris", JSAP.NO_SHORTFLAG, "osiris",
                "write to OSIRIS format"));
        jsap.registerParameter(new Switch("gz", 'z', "gzip",
                "force gzip compression for SOMLib- and ARFF-Outputfiles."));

        // Misc
        jsap.registerParameter(new Switch("nogui", JSAP.NO_SHORTFLAG, "nogui",
                "disable all graphical user interfaces (like progress windows, etc.)"));
        jsap.registerParameter(new FlaggedOption("threads", JSAP.INTEGER_PARSER, "1", false, 't',
                "threads", "specify number of parallel extraction threads (default = 1)"));
    }

    private int parallelThreads = 1;

    private ExtractionDispatcher dispatcher = null;

    private static void checkFeatureExtractionOptions(FeatureExtractionOptions opt) {
        if (opt.inputBaseDir == null) {
            opt.inputBaseDir = "";
        }

        if (opt.inputBaseDir.length() > 0) {
            if (!opt.inputBaseDir.endsWith(File.separator)) {
                opt.inputBaseDir = opt.inputBaseDir + File.separator;
            }
        }
    }

    /**
     * extract files recursively from directory, search for standard audio extensions
     * 
     * @param file directory, or, if only 1 single file is to be extracted: file
     * @param opt Feature Extraction options
     * @return a list of {@link FeatureVectorData} containing the extracted features.
     */
    public static List<FeatureVectorData> createExtractionList(File file,
            FeatureExtractionOptions opt) {
        return createExtractionList(file, DEFAULT_AUDIO_EXTENSIONS, opt);
    }

    /**
     * extract files recursively from directory, search for defined audio extensions provided as
     * parameter
     * 
     * @param file directory, or, if only 1 single file is to be extracted: file
     * @param fileExtensions a string array containing file extensions for which should be searched
     *            while recursing directories
     * @param opt Feature Extraction options
     * @return a list of {@link FeatureVectorData} containing the extracted features.
     */
    public static List<FeatureVectorData> createExtractionList(File file,
            final String[] fileExtensions, FeatureExtractionOptions opt) {
        List<File> fileList = new LinkedList<File>();

        if (!file.exists()) {
            // TODO throw exception
            MyLogger.severe("File or directory \"" + file.getPath() + "\" does not exist!");
            return new LinkedList<FeatureVectorData>();
        }

        if (file.isDirectory()) {
            opt.inputBaseDir = file.getAbsolutePath();
            // System.out.println("IN: " + opt.inputBaseDir);

            FileFilter extensionFilter = new FileFilter() {
                @Override
                public boolean accept(File pathname) {
                    // Accept directories
                    if (pathname.isDirectory()) {
                        return true;
                    }

                    for (String fileExtension : fileExtensions) {
                        String suffix = pathname.getName().substring(
                                pathname.getName().length() - fileExtension.length());
                        if (fileExtension.equalsIgnoreCase(suffix)) {
                            return true;
                        }
                    }

                    return false;
                }
            };

            long bytesTotal = getFileList(file, fileList, extensionFilter);

            int numFiles = fileList.size();
            MyLogger.info(numFiles + " Files. Total Bytes: " + bytesTotal);
        } else // treat single file
        {
            fileList.add(file);
        }

        /* create metadata list */

        if (opt.inputBaseDir.length() > 0) {
            opt.inputBaseDir = new File(opt.inputBaseDir).getAbsolutePath() + File.separator;
        }

        Iterator<File> it = fileList.iterator();
        List<FeatureVectorData> metaDataList = new LinkedList<FeatureVectorData>();
        FeatureVectorData featVecMetaData;
        while (it.hasNext()) /* iterate files */
        {
            featVecMetaData = new FeatureVectorData();
            featVecMetaData.setFileLabel(getRelativeFilename(it.next(), opt.inputBaseDir));
            metaDataList.add(featVecMetaData);
        }

        // extractListOfFiles(metaDataList, opt);
        return metaDataList;
    }

    /**
     * cuts base directory out of absolute filename
     * 
     * @param file File object of a file, given with absolute filename
     * @param baseDir String containing base directory, with file separator at the end
     * @return relative filename
     */
    public static String getRelativeFilename(File file, String baseDir) {
        if (baseDir != null) {
            if (baseDir.length() > 0) {
                return file.getAbsolutePath().substring(baseDir.length());
            }
        }

        return file.getPath();
    }

    /**
     * adds base directory at the beginning of relative filenames, only if base directory is
     * non-null and not empty string
     * 
     * @param filename String containing relative filename
     * @param baseDir String containing base directory, with file separator at the end
     * @return String containing absolute filename (if base directory is set)
     */
    public static String getAbsoluteFilename(String filename, String baseDir) {
        if (baseDir != null) {
            if (baseDir.length() > 0) {
                return baseDir + filename;
            }
        }
        return filename;
    }

    /**
     * extract features from a list of files provided from an Ascii text file containing one
     * filename per line
     * 
     * @param fileListfile file containing one filename per line
     * @param opt Feature Extraction options
     */
    // public void extractListOfFiles(File fileListfile,
    // FeatureExtractionOptions opt)
    // {
    //
    //
    // extractListOfFiles(list, opt);
    // }
    /**
     * extract features from a list of files provided by a Java List object
     * 
     * @param fileList List object containing File objects
     * @param opt Feature Extraction options
     */

    // public static void extractListOfFiles(List fileList,
    // FeatureExtractionOptions opt)
    // {
    //
    // }
    /**
     * extract features from a list of files provided from an Ascii text file containing one
     * filename per line
     * 
     * @param fileListfile file containing one filename per line
     * @param opt Feature Extraction options
     */
    // public void extractListOfFiles(File fileListfile,
    // FeatureExtractionOptions opt)
    // {
    //
    //
    // extractListOfFiles(list, opt);
    // }
    /**
     * extract features from a list of files provided by a Java List object
     * 
     * @param fileList List object containing File objects
     * @param opt Feature Extraction options
     */

    // public static void extractListOfFiles(List fileList,
    // FeatureExtractionOptions opt)
    // {
    //
    // }
    /**
     * extract features from a list of files provided by a Java List object
     * 
     * @param featureVectorMetaDataList List object containing FeatureVectorData objects
     * @param opt Feature Extraction options
     * @param graphic reference to the main gui, may be <code>null</code>
     */
    public void extractListOfFiles(List<FeatureVectorData> featureVectorMetaDataList,
            FeatureExtractionOptions opt, MainGUI graphic) {
        checkFeatureExtractionOptions(opt);

        int numFiles = featureVectorMetaDataList.size();
        long bytesTotal = checkFiles(featureVectorMetaDataList, opt.inputBaseDir);

        if (graphic != null) {
            graphic.setFeatureExtractionOptions(opt);
        }

        @SuppressWarnings("unused")
        long bytesDone = 0;

        if (numFiles == 0) {
            MyLogger.error("0 files found. Aborting.");
            if (graphic != null) {
                graphic.setStatus(Status.SET_UP);
            }
            return;
        }

        DateFormat df = DateFormat.getTimeInstance(DateFormat.DEFAULT);
        df.setTimeZone(TimeZone.getTimeZone("Etc/UCT"));
        MyLogger.info("STARTING FEATURE EXTRACTION at " + df.format(new Date()));
        MyLogger.info(numFiles + " files, " + bytesTotal + " bytes.");

        int numFeatSets = opt.getNumberOfFeatureSets();
        int numOutputFormats = opt.getNumberOfOutputFormats();
        String[] extensions = opt.getFileExtensionsOfFeatureSets();
        AbstractDataWriter[][] outputFiles = null;

        if (numOutputFormats > 0 && numFeatSets > 0) {

            outputFiles = new AbstractDataWriter[numFeatSets][numOutputFormats];

            for (int feat = 0; feat < extensions.length; feat++) {
                int file = 0;
                if (opt.writeSOMLib) {
                    try {
                        outputFiles[feat][file] = new SOMLibFormatWriter(opt.outputBaseFilename
                                + "." + extensions[feat] + ".vec", opt.enableSOMLibGZip);
                        outputFiles[feat][file].setDataType("audio-" + extensions[feat]);
                        if (opt.inputBaseDir.length() > 0) {
                            outputFiles[feat][file].setBaseDirectory(opt.inputBaseDir);
                        }
                        file++;
                    } catch (IOException e) {
                        MyLogger.severe("An error occured while opening SOMLib vector file to write: "
                                + e.getMessage());
                        if (graphic != null) {
                            graphic.setStatus(Status.SET_UP);
                        }

                        return;
                    }
                }
                if (opt.writeXML) {
                    try {
                        outputFiles[feat][file] = new XMLFormatWriter(opt.outputBaseFilename + "."
                                + extensions[feat] + ".xml");
                        outputFiles[feat][file].setDataType("audio-" + extensions[feat]);
                        if (opt.inputBaseDir.length() > 0) {
                            outputFiles[feat][file].setBaseDirectory(opt.inputBaseDir);
                        }
                        file++;
                    } catch (IOException e) {
                        MyLogger.severe("An error occured while opening XML vector file to write: "
                                + e.getMessage());
                        if (graphic != null) {
                            graphic.setStatus(Status.SET_UP);
                        }

                        return;
                    }
                }
                if (opt.writeOSIRIS) {
                    try {
                        outputFiles[feat][file] = new OSIRISFormatWriter(opt.outputBaseFilename
                                + "." + extensions[feat] + ".osiris");
                        outputFiles[feat][file].setDataType("audio-" + extensions[feat]);
                        if (opt.inputBaseDir.length() > 0) {
                            outputFiles[feat][file].setBaseDirectory(opt.inputBaseDir);
                        }
                        file++;
                    } catch (IOException e) {
                        MyLogger.severe("An error occured while opening OSIRIS vector file to write: "
                                + e.getMessage());
                        return;
                    }
                }
                if (opt.writeWekaARFF) {
                    try {
                        outputFiles[feat][file] = new ARFFFormatWriter(opt.outputBaseFilename + "."
                                + extensions[feat] + ".arff", opt.enableArffGZip,
                                opt.enableArffVectorLabel);
                        outputFiles[feat][file].setDataType("audio-" + extensions[feat]);
                        if (opt.inputBaseDir.length() > 0) {
                            outputFiles[feat][file].setBaseDirectory(opt.inputBaseDir);
                        }
                        file++;
                    } catch (IOException e) {
                        MyLogger.severe("An error occured while opening ARFF vector file to write: "
                                + e.getMessage());
                        if (graphic != null) {
                            graphic.setStatus(Status.SET_UP);
                        }

                        return;
                    }
                }
            }
        }
        dispatcher = new ExtractionDispatcher(featureVectorMetaDataList, bytesTotal, opt,
                outputFiles, false, graphic);

        // This method blocks until extraction is done
        dispatcher.extract(parallelThreads);

        int numErrors = dispatcher.numErrors;

        // TODO inline

        if (numErrors > 0) {
            MyLogger.info(numErrors + " ERRORS occured:");
            MyLogger.info(dispatcher.errorLog.toString());
            MyLogger.info("");
        }

        long elapsedTime = System.currentTimeMillis() - dispatcher.startTime;
        String remainInfo = "FINISHED. Total duration: "
                + at.tuwien.ifs.somtoolbox.util.DateUtils.shortFormatDuration(elapsedTime);
        MyLogger.info(remainInfo);
        dispatcher = null;
    }

    public static long checkFiles(List<FeatureVectorData> fileList, String baseDir) {
        long bytes = 0;

        Iterator<FeatureVectorData> it = fileList.iterator();

        while (it.hasNext()) {
            // error = false;
            FeatureVectorData curFeatVecMetaData = it.next();
            File curFile = new File(getAbsoluteFilename(curFeatVecMetaData.getFileLabel(), baseDir));
            // TODO check if exists
            // TODO check if format is readable
            // TODO check sampleRate etc.
            bytes += curFile.length();
        }

        return bytes;
    }

    /**
     * constructs a FileList from recursing a directory for files with a given list of extensions
     * 
     * @param file directory to start recursion
     * @param result resulting List of files found
     * @param filter a file filter used to determine the files to include in the list.
     * @return the sum of the size of all files found in bytes
     */
    public static long getFileList(File file, List<File> result, FileFilter filter) {
        if (file.isDirectory()) {
            long total_length = 0;
            File[] child = file.listFiles(filter);
            if (child != null && child.length > 0) {
                for (File element : child) {
                    total_length += getFileList(element, result, filter);
                }
            }
            return total_length;
        } else {
            // check for valid extensions
            result.add(file);
            // System.out.println(file.toString() + " " +
            // file.length());
            return file.length();
        }
    }

    public int getParallelThreads() {
        return parallelThreads;
    }

    public void setParallelThreads(int parallelThreads) {
        this.parallelThreads = parallelThreads;
        if (dispatcher != null) {
            dispatcher.setDesiredThreadCount(parallelThreads);
        }
    }

    public void cancelExtraction() {
        if (dispatcher != null) {
            dispatcher.cancelExtraction();
        }
    }
}
