/*
 * Copyright 2004-2010 Information & Software Engineering Group (188/1)
 *                     Institute of Software Technology and Interactive Systems
 *                     Vienna University of Technology, Austria
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.ifs.tuwien.ac.at/mir/index.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package at.tuwien.ifs.feature.extraction.audio.distance;

//import at.ec3.somtoolbox.data.InputDatum;
//import cern.colt.matrix.DoubleMatrix1D;

/**
 * Implements a static method for metric instantiation and a method for mean vector calculation.
 * Furthermore, implements the convenience methods to call the abstract method having two double
 * arrays as arguments (see {@link #distance(double[], double[])}). This method has to be
 * implemented by classes actually implementing a certain metric.
 * 
 * @author mbach
 * @version $Id: AbstractMetric.java 167 2010-06-17 10:12:01Z frank $
 */
public abstract class AbstractMetric implements DistanceMetric {

    /**
     * Instantiates a certain distance metric class specified by argument <code>mName</code>.
     * 
     * @param mName the name of the metric.
     * @return a distance metric object of class <code>mName</code>.
     * @throws ClassNotFoundException if class denoted by argument <code>mName</code> is not found.
     * @throws InstantiationException if if this Class represents an abstract class, an interface,
     *             an array class, a primitive type, or void; or if the class has no nullary
     *             constructor; or if the instantiation fails for some other reason.
     * @throws IllegalAccessException if the class or its nullary constructor is not accessible.
     */
    public static DistanceMetric instantiate(String mName) throws ClassNotFoundException,
            InstantiationException, IllegalAccessException {
        DistanceMetric metric = null;
        Class<?> c = Class.forName(mName);
        metric = (DistanceMetric) c.newInstance();
        return metric;
    }

    /**
     * Calculates the mean vector of two double array vectors.
     * 
     * @param vector1 first vector.
     * @param vector2 second vector.
     * @return the mean vector.
     * @throws MetricException if the dimensionalities of the two vectors differ.
     */
    public static double[] meanVector(double[] vector1, double[] vector2) throws MetricException {
        if (vector1.length != vector2.length) {
            throw new MetricException(
                    "Oops ... tried to calculate the mean vector of two vectors with different dimensionalities.");
        }
        // initialize mean vector
        double[] meanVector = new double[vector1.length];
        for (int ve = 0; ve < vector1.length; ve++) { // calculating mean vector
            meanVector[ve] = (vector1[ve] + vector2[ve]) / (double) 2;
        }
        return meanVector;
    }

    /**
     * @see DistanceMetric#distance(double[], double[])
     */
    public abstract double distance(double[] vector1, double[] vector2) throws MetricException;

    // /**
    // * @see at.ec3.somtoolbox.layers.metrics.DistanceMetric#distance(double[],
    // cern.colt.matrix.DoubleMatrix1D)
    // */
    // public double distance(double[] vector1, DoubleMatrix1D vector2) throws
    // MetricException {
    // return distance(vector1,vector2.toArray());
    // }
    //
    // /**
    // * @see at.ec3.somtoolbox.layers.metrics.DistanceMetric#distance(double[],
    // at.ec3.somtoolbox.data.InputDatum)
    // */
    // public double distance(double[] vector, InputDatum data) throws
    // MetricException {
    // return distance(vector, data.getVector());
    // }
    //
    // /**
    // * @see
    // at.ec3.somtoolbox.layers.metrics.DistanceMetric#distance(cern.colt.matrix.DoubleMatrix1D,
    // double[])
    // */
    // public double distance(DoubleMatrix1D vector1, double[] vector2) throws
    // MetricException {
    // return distance(vector1.toArray(),vector2);
    // }
    //    
    // /**
    // * @see
    // at.ec3.somtoolbox.layers.metrics.DistanceMetric#distance(cern.colt.matrix.DoubleMatrix1D,
    // cern.colt.matrix.DoubleMatrix1D)
    // */
    // public double distance(DoubleMatrix1D vector1, DoubleMatrix1D vector2)
    // throws MetricException {
    // return distance(vector1.toArray(),vector2.toArray());
    // }
    //
    // /**
    // * @see
    // at.ec3.somtoolbox.layers.metrics.DistanceMetric#distance(cern.colt.matrix.DoubleMatrix1D,
    // at.ec3.somtoolbox.data.InputDatum)
    // */
    // public double distance(DoubleMatrix1D vector, InputDatum datum) throws
    // MetricException {
    // return distance(vector, datum.getVector());
    // }
    //
    // /**
    // * @see
    // at.ec3.somtoolbox.layers.metrics.DistanceMetric#distance(at.ec3.somtoolbox.data.InputDatum,
    // double[])
    // */
    // public double distance(InputDatum data, double[] vector) throws
    // MetricException {
    // return distance(data.getVector(), vector);
    // }
    //    
    // /**
    // * @see
    // at.ec3.somtoolbox.layers.metrics.DistanceMetric#distance(at.ec3.somtoolbox.data.InputDatum,
    // cern.colt.matrix.DoubleMatrix1D)
    // */
    // public double distance(InputDatum datum, DoubleMatrix1D vector) throws
    // MetricException {
    // return distance(datum.getVector(), vector);
    // }

    /**
     * Performs a check on wether the given vectors have the same dimension.
     * 
     * @throws MetricException If the given vectors have different dimensions.
     */
    protected void checkDimensions(double[] vector1, double[] vector2) throws MetricException {
        if (vector1.length != vector2.length) {
            throw new MetricException(
                    "Tried to calculate distance between two vectors with different dimensionalities.");
        }
    }

    /**
     * Performs a check on wether the given vectors have the same dimension.
     * 
     * @throws MetricException If the given vectors have different dimensions.
     */
    /*
     * protected void checkDimensions(DoubleMatrix1D vector1, DoubleMatrix1D vector2) throws
     * MetricException { if (vector1.cardinality() != vector2.cardinality()) { throw new
     * MetricException(
     * "Tried to calculate distance between two vectors with different dimensionalities." ); } }
     */

    /*
     * public double transformValue(double value) { return value; } public double[]
     * transformVector(double[] vector) { double[] transformed = new double[vector.length]; for (int
     * i = 0; i < transformed.length; i++) { transformed[i] = transformValue(vector[i]); } return
     * transformed; }
     */

}
