/*
 * Copyright 2004-2010 Information & Software Engineering Group (188/1)
 *                     Institute of Software Technology and Interactive Systems
 *                     Vienna University of Technology, Austria
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.ifs.tuwien.ac.at/mir/index.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package at.tuwien.ifs.feature.extraction.audio.input;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Map;

import javax.sound.sampled.AudioFileFormat;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.UnsupportedAudioFileException;

import org.tritonus.share.sampled.TAudioFormat;
import org.tritonus.share.sampled.file.TAudioFileFormat;

import at.tuwien.ifs.feature.extraction.audio.util.MyLogger;

/**
 * Copyright Vienna University of Technology
 * 
 * @author Thomas Lidy
 * @author Jakob Frank
 * @version $Id: AudioFileSegmentReader.java 173 2010-06-17 16:28:05Z mayer $
 */
public class AudioFileSegmentReader {

    // private File file;
    private InputStream inStream = null;

    private AudioInputStream audioInputStream = null;

    private AudioInputStream mpegInputStream = null; // needed if reading mp3s

    private AudioFileFormat audioFileFormat = null;

    private AudioFormat audioFormat = null;

    private boolean open = false;

    private boolean isMpeg = false;

    private int sampleRate;

    private int sampleSizeInBits;

    private int nChannels;

    // private int frameSize;
    @SuppressWarnings("unused")
    private long lengthInFrames;

    private long durationMicrosec = 0;

    private long streamLength;

    @SuppressWarnings("unused")
    private static byte[] convertFileToByteArray(File f) throws IOException {
        FileInputStream fis = new FileInputStream(f);
        byte[] tmp = new byte[fis.available()];
        fis.read(tmp);
        fis.close();
        return tmp;
    }

    public AudioFileSegmentReader(File file) throws UnsupportedAudioFileException, IOException {
        // this((new FileInputStream(file)));
        this(AudioSystem.getAudioInputStream(file));
    }

    public AudioFileSegmentReader(AudioInputStream inStream) throws UnsupportedAudioFileException,
            IOException {
        this.inStream = inStream;
        this.streamLength = inStream.available();
        audioFileFormat = AudioSystem.getAudioFileFormat(inStream);
        audioFormat = audioFileFormat.getFormat();
        /*
         * static AudioFileFormat AudioSystem.getAudioFileFormat (java.io.File file) static
         * AudioFileFormat getAudioFileFormat(java.io.InputStream stream) static AudioFileFormat
         * getAudioFileFormat (java.net.URL url)
         */
        MyLogger.info(audioFormat.toString());
        // System.out.println(audioFileFormat.getType().toString());
        isMpeg = audioFormat.getEncoding().toString().startsWith("MPEG");

        openStream();

        // frameSize = audioInputStream.getFormat().getFrameSize();
        // is sampleSizeInBits/8 * nChannels in WAV & AU files
        // but much higher in MPEG files

        lengthInFrames = audioInputStream.getFrameLength();

        audioFormat = audioInputStream.getFormat(); // needs to be read again in
        // case of mp3s (for wav
        // layer values)
        sampleRate = (int) audioFormat.getSampleRate();
        sampleSizeInBits = audioFormat.getSampleSizeInBits();
        nChannels = audioFormat.getChannels();

        if ((isMpeg) && (audioFileFormat instanceof TAudioFileFormat)) {
            Map<?, ?> properties = ((TAudioFileFormat) audioFileFormat).properties();
            // durationMicrosec = ((Long)
            // properties.get("duration")).longValue();
            long bitrate = ((Integer) properties.get("mp3.bitrate.nominal.bps")).longValue();
            durationMicrosec = ((streamLength * 8L) / bitrate) * 1000000L;

            MyLogger.extrainfo("duration (microsec) [mpeg]: " + durationMicrosec);
            // Integer val = (Integer) properties.get("mp3.length.frames");
            // MyLogger.extrainfo("mp3.length.frames: " + val);
        } else {
            durationMicrosec = audioInputStream.getFrameLength() * 1000 * 1000 / sampleRate;
            MyLogger.extrainfo("duration (microsec) [norm]: " + durationMicrosec);
        }

    }

    public int getSampleRate() {
        return sampleRate;
    }

    public int getSampleSizeInBits() {
        return sampleSizeInBits;
    }

    public long getNumberOfSamples() {
        if (isMpeg) {
            // audioInputStream.getFrameLength() returns -1 for MP3s
            // thus another method is needed:
            // the following WORKAROUND will approximate the number of samples
            // only
            // - usually it is SMALLER!

            return durationMicrosec / 1000 * (long) sampleRate / 1000;
        } else
            return audioInputStream.getFrameLength();
    }

    public int getDurationInSeconds() {
        return (int) (durationMicrosec / 1000 / 1000);
    }

    @SuppressWarnings("unused")
    private void readMpegProperties() {
        if (audioFileFormat instanceof TAudioFileFormat) {
            Map<?, ?> properties = ((TAudioFileFormat) audioFileFormat).properties();
            String key = "mp3.length.bytes";
            Integer val = (Integer) properties.get(key);
            System.out.println(key + ": " + val);
            key = "mp3.length.frames";
            val = (Integer) properties.get(key);
            System.out.println(key + ": " + val);
            // key = "mp3.id3tag.v2";
            // InputStream tag= (InputStream) properties.get(key);
            System.out.println(properties.toString());
        }

        if (audioFormat instanceof TAudioFormat) {
            Map<?, ?> properties = ((TAudioFormat) audioFormat).properties();
            String key = "bitrate";
            Integer val = (Integer) properties.get(key);
            System.out.println(key + ": " + val);
            System.out.println(properties.toString());
        }
    }

    public void openStream() throws UnsupportedAudioFileException, IOException {
        if (isMpeg) {
            mpegInputStream = AudioSystem.getAudioInputStream(inStream);
            AudioFormat baseFormat = mpegInputStream.getFormat();
            AudioFormat decodedFormat = new AudioFormat(AudioFormat.Encoding.PCM_SIGNED,
                    baseFormat.getSampleRate(), 16, baseFormat.getChannels(),
                    baseFormat.getChannels() * 2, baseFormat.getSampleRate(), false);
            audioInputStream = AudioSystem.getAudioInputStream(decodedFormat, mpegInputStream);
            // System.out.println("ismpeg");
        } else {
            audioInputStream = AudioSystem.getAudioInputStream(inStream);
        }

        open = true;
    }

    public void resetStream() throws IOException {
        if (audioInputStream != null) {
            audioInputStream.reset();
        }
    }

    public void closeStream() throws IOException {
        if (audioInputStream != null) {
            audioInputStream.close();
        }
        if (mpegInputStream != null) {
            mpegInputStream.close();
        }

        open = false;
    }

    public int readMono(short[] outputArray) throws IOException {
        if (nChannels == 1) {
            return read(outputArray);
        } else {
            int len = outputArray.length;
            short[] stereoArray = new short[len * 2];
            int bytesRead = read(stereoArray);
            // TODO: ensure thate array elements > bytesRead are set to 0
            for (int i = 0; i < len; i++) {
                outputArray[i] = (short) ((short) (stereoArray[i * 2] + stereoArray[i * 2 + 1]) / 2);
            }

            if (bytesRead == -1)
                return -1;
            return bytesRead / 2;
        }
    }

    public int read(short[] outputArray) throws IOException {
        int len = outputArray.length;
        int bytesRead;

        /* 16 bit or 8 bit files */
        if (sampleSizeInBits == 16) {
            byte[] byteArray = new byte[len * 2];
            bytesRead = read(byteArray);
            // System.out.println("bytesRead: " + bytesRead);

            if (audioFormat.isBigEndian()) // e.g. AU files
            {
                for (int i = 0; i < len; i++) {
                    outputArray[i] = (short) ((byteArray[i * 2 + 1] & 0xff) | ((byteArray[i * 2] & 0xff) << 8));
                }
            } else // WAV files
            {
                for (int i = 0; i < len; i++) {
                    outputArray[i] = (short) (((byteArray[i * 2 + 1] & 0xff) << 8) | (byteArray[i * 2] & 0xff));
                }
            }

            // TODO: ensure thate array elements > bytesRead are set to 0

            if (bytesRead == -1)
                return -1;
            return bytesRead / 2;
        } else {
            byte[] byteArray = new byte[len];
            bytesRead = read(byteArray);
            // TODO: ensure thate array elements > bytesRead are set to 0
            for (int i = 0; i < len; i++) {
                outputArray[i] = (short) byteArray[i];
            }
            return bytesRead;
        }
    }

    /**
     * @param outputArray array into which input is read (# of bytes to read equals size of array)
     * @return number of bytes actually read (can be smaller at the end of file, or -1 in case of
     *         errors)
     * @throws IOException
     */
    public int read(byte[] outputArray) throws IOException {
        if (!open) {
            try {
                openStream();
            } catch (UnsupportedAudioFileException e) {
                // this should not happen, because openStream is called in
                // constructor already
                // and an exception should be noticed already there
                e.printStackTrace();
            }
        }

        int len = outputArray.length;
        int bytesDone = 0;
        int bytesToRead = len;
        int bytesRead = 0;

        /* read bytes as long as you can */
        /*
         * (in mp3s, maximum #bytes read at once = framesize, thus we read multiple times
         */

        while (bytesToRead > 0) {
            // Syntax: read(intoArray, fromOffset, numberOfBytes)
            bytesRead = audioInputStream.read(outputArray, bytesDone, bytesToRead);

            if (bytesRead == -1)
                break;

            bytesDone += bytesRead;
            bytesToRead -= bytesRead;
        }

        if (bytesDone > 0)
            return bytesDone;
        else
            return -1;
    }

    /**
     * Convenience method for getting a 16-bit sample from a byte array. Samples should be in
     * 16-bit, signed, little-endian format. private short getSample(byte[] buffer, int position) {
     * return (short) (((buffer[position + 1] & 0xff) << 8) | (buffer[position] & 0xff)); }
     */

}
