/*
 * Copyright 2004-2010 Information & Software Engineering Group (188/1)
 *                     Institute of Software Technology and Interactive Systems
 *                     Vienna University of Technology, Austria
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.ifs.tuwien.ac.at/mir/index.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package at.tuwien.ifs.feature.extraction.audio.output;

import java.io.IOException;
import java.util.ArrayList;

import at.tuwien.ifs.feature.extraction.audio.data.FeatureVectorData;
import at.tuwien.ifs.somtoolbox.data.InputDataWriter;
import at.tuwien.ifs.somtoolbox.util.StringUtils;

/**
 * Copyright Vienna University of Technology
 * 
 * @author Rudolf Mayer
 * @version $Id: SOMLibFormatWriter.java 173 2010-06-17 16:28:05Z mayer $
 */
public class ARFFFormatWriter extends AbstractDataWriter {

    private final String realFileName;

    private int numVectors = 0;

    private boolean writeLabels = false;

    private ArrayList<StringBuilder> tmpVectors = new ArrayList<StringBuilder>();

    private ArrayList<String> tmpLabels = new ArrayList<String>();

    /**
     * Constructor for ARFFFormatWriter. if the outFileName ends with ".gz" gzip compression will be
     * enabled.
     * 
     * @param outFileName the output filename
     * @param gzipped set <code>true</code> to force gzip compression
     * @throws IOException
     */
    public ARFFFormatWriter(String outFileName, boolean gzipped, boolean writeLabels)
            throws IOException {
        super(outFileName);
        this.writeLabels = writeLabels;
        realFileName = outFileName;
    }

    private void writeHeader() throws IOException {
        outFile.write("% Created by Java Audio Feature Extraction");
        outFile.newLine();
        outFile.write("%   (c) Institute of Software Technology and Interactive Systems");
        outFile.newLine();
        outFile.write("%       Vienna University of Technology, Austria");
        outFile.newLine();
        outFile.newLine();
        outFile.write("@RELATION \"" + realFileName + "\"");
        outFile.newLine();
        for (int i = 0; i < featureVectorLength; i++) {
            outFile.write("@ATTRIBUTE feature" + (i + 1) + "\tNUMERIC");
            outFile.newLine();
        }

        // write instance name attribute?
        if (writeLabels) {
            outFile.write("@ATTRIBUTE instanceName string");
            outFile.newLine();
        }

        outFile.write(InputDataWriter.getWekaClassHeader(classAssignmentResolver.getClassNames().toArray(
                new String[classAssignmentResolver.getClassNames().size()])));

        outFile.newLine();
        outFile.newLine();
    }

    private void writeData() throws IOException {
        outFile.write("@DATA");
        outFile.newLine();
        for (int i = 0; i < numVectors; i++) {
            try {
                outFile.write(tmpVectors.get(i).toString());

                // write class
                outFile.write("'"
                        + StringUtils.escapeClassNameForWeka(classAssignmentResolver.getClassAssignment(tmpLabels.get(i)))
                        + "'");

                outFile.newLine();
            } catch (IOException e) {
                throw new IOException("Problem writing ARFF vector file " + fileName + "! "
                        + e.getMessage());
            }
        }
    }

    @Override
    public void writeVector(FeatureVectorData featureVector) throws IOException {
        double[][] matrix = featureVector.getMatrixData();
        int vectorDim = matrix.length * matrix[0].length;
        checkVectorDim(vectorDim);
        StringBuilder sb = new StringBuilder();
        for (int c = 0; c < matrix[0].length; c++) {
            for (double[] element : matrix) {
                sb.append(formatDouble.format(element[c])).append(",");
            }
        }

        // write instance name attribute?
        if (writeLabels) {
            sb.append("'" + StringUtils.escapeForWeka(featureVector.getFileLabel()) + "',").append(
                    ",");
        }

        tmpVectors.add(sb);
        tmpLabels.add(featureVector.getFileLabel());
    }

    @Override
    public void close() throws IOException {
        numVectors = tmpVectors.size();
        writeHeader();
        writeData();

        try {
            super.close();
        } catch (IOException e) {
            throw new IOException("Problem closing ARFF vector file " + fileName + "! "
                    + e.getMessage());
        }

    }

}
