/*
 * Copyright 2004-2010 Information & Software Engineering Group (188/1)
 *                     Institute of Software Technology and Interactive Systems
 *                     Vienna University of Technology, Austria
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.ifs.tuwien.ac.at/mir/index.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package at.tuwien.ifs.feature.extraction.audio.output;

import java.io.BufferedWriter;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.zip.GZIPOutputStream;

import at.tuwien.ifs.feature.extraction.audio.data.FeatureVectorData;

public abstract class AbstractDataWriter {

    protected final String fileName;

    protected final BufferedWriter outFile;

    protected String nl = "\n"; // new line character(s) - can be fixed for all

    // systems, or System property "line.separator"

    protected int featureVectorLength = 0;

    protected String dataTypeID = null;

    protected String baseDirectory = null;

    protected DecimalFormat formatFloat = null;

    protected DecimalFormat formatDouble = null;

    protected boolean gzipped = false;

    protected ClassAssignmentResolver classAssignmentResolver = null;

    /* PUBLIC METHODS */

    public AbstractDataWriter(String outFileName) throws IOException {
        this(outFileName, false);
    }

    public AbstractDataWriter(String outFileName, boolean gzipped) throws IOException {
        this.gzipped = gzipped;

        if (gzipped) {
            if (!outFileName.endsWith(".gz")) {
                outFileName = outFileName + ".gz";
            }
        }

        fileName = outFileName;

        if (gzipped) {
            outFile = new BufferedWriter(new OutputStreamWriter(new GZIPOutputStream(
                    new FileOutputStream(fileName))));
        } else {
            outFile = new BufferedWriter(new FileWriter(fileName));
        }

        initDecimalFormats();
    }

    public void close() throws IOException {
        try {
            outFile.close();
        } catch (IOException e) {
            throw new IOException("Problem closing file " + fileName + "! " + e.getMessage());
        }
    }

    /**
     * If data type is set, e.g. the header line $DATA_TYPE will be included in the SOMLib vector
     * file.
     * 
     * @param dataType set data type ID, e.g. "audio-rp"
     */
    public void setDataType(String dataType) {
        dataTypeID = dataType;
    }

    /**
     * If the base directory is provided, it will be included in the header of the output files.
     * 
     * @param baseDir base directory of files provided as fileLabel to the vectors
     */
    public void setBaseDirectory(String baseDir) {
        baseDirectory = baseDir;
    }

    /* THE FOLLOWING METHODS *MUST* BE OVERRIDDEN BY SUBCLASSES */

    public abstract void writeVector(FeatureVectorData featureVector) throws IOException;

    /* PRIVATE/PROTECTED METHODS */

    protected void checkVectorDim(int vectorDim) throws IOException {
        if (featureVectorLength > 0 && featureVectorLength != vectorDim) {
            throw new IOException(
                    "Cannot write feature vectors with different lengths into one file! "
                            + "Previous vector length was " + featureVectorLength
                            + " current one is " + vectorDim + "!");
        } else {
            featureVectorLength = vectorDim;
        }
    }

    protected void initDecimalFormats() {
        // DecimalFormat form = new DecimalFormat("#.###############");
        // Locale def = Locale.getDefault();
        // System.out.println("LOCALE: " + def.toString());

        NumberFormat form1 = NumberFormat.getInstance(java.util.Locale.US);
        // avoids format troubles on localized OS's (, instead of .)

        if (form1 instanceof DecimalFormat) {
            formatDouble = (DecimalFormat) form1;
            formatDouble.applyPattern("#.###############");
            // System.out.print("NF: " + ((DecimalFormat) form).toPattern());
        } else {
            String errorMessage = "NumberFormat initialization failed. Aborting.";
            throw new ClassCastException(errorMessage);
        }

        form1 = NumberFormat.getInstance(java.util.Locale.US);
        // avoids format troubles on localized OS's (, instead of .)

        if (form1 instanceof DecimalFormat) {
            formatFloat = (DecimalFormat) form1;
            formatFloat.applyPattern("#.#####");
            // System.out.print("NF: " + ((DecimalFormat) form).toPattern());
        } else {
            String errorMessage = "NumberFormat initialization failed. Aborting.";
            throw new ClassCastException(errorMessage);
        }
    }

    public void setClassAssignmentResolver(ClassAssignmentResolver classAssignmentResolver) {
        this.classAssignmentResolver = classAssignmentResolver;
    }
}
