/*
 * Copyright 2004-2010 Information & Software Engineering Group (188/1)
 *                     Institute of Software Technology and Interactive Systems
 *                     Vienna University of Technology, Austria
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.ifs.tuwien.ac.at/mir/index.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package at.tuwien.ifs.feature.extraction.audio.output;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

import at.tuwien.ifs.somtoolbox.util.StringUtils;

/**
 * This class stores label names of all the inputs, and finally resolves class names and assignments
 * of the labels to classes. Currently, the assignment is simply done by
 * <ul>
 * <li>removing the prefix common to all labels, and then</li>
 * <li>taking the directory part of each label name as class</li>
 * <li>labels that have no directory element in their name will be assigned to the class
 * {@link #UNKNOWN}</li>
 * </ul>
 * 
 * @author Rudolf Mayer
 * @version $Id: $
 */
public class ClassAssignmentResolver {
    private static final String UNKNOWN = "Unknown";

    private ArrayList<String> labels = new ArrayList<String>();

    private HashMap<String, String> classAssignments = new HashMap<String, String>();

    private HashSet<String> classNames = new HashSet<String>();

    public void addLabel(String label) {
        labels.add(label);
    }

    public void resolveClassAssignment() {
        // find the class assignment

        // step 1: find the longest common substring
        String prefix = StringUtils.getCommonPrefix(labels);
        int prefixLength = prefix.length();

        // step 2: find the different class names, by looking at subdirectories
        for (int i = 0; i < labels.size(); i++) {
            String label = labels.get(i);
            String suffix = label.substring(prefixLength);
            if (suffix.contains(File.separator)) {
                String directoryName = suffix.substring(0, suffix.lastIndexOf(File.separator));
                classNames.add(directoryName);
                classAssignments.put(label, directoryName);
            } else { // if this object doesn't have a path, we put it into a generic class
                classNames.add(UNKNOWN);
                classAssignments.put(label, UNKNOWN);
            }
        }

    }

    public HashMap<String, String> getClassAssignments() {
        return classAssignments;
    }

    public String getClassAssignment(String label) {
        String className = classAssignments.get(label);
        if (className == null) {
            if (labels.contains(label)) {
                throw new IllegalArgumentException("Label '" + label + "' is not known.");
            } else {
                throw new IllegalArgumentException("Class assignment for label '" + label
                        + "' was not successful.");
            }
        } else {
            return className;
        }
    }

    public HashSet<String> getClassNames() {
        return classNames;
    }
}
