/*
 * Copyright 2004-2010 Information & Software Engineering Group (188/1)
 *                     Institute of Software Technology and Interactive Systems
 *                     Vienna University of Technology, Austria
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.ifs.tuwien.ac.at/mir/index.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package at.tuwien.ifs.feature.extraction.audio.output;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.zip.GZIPOutputStream;

import at.tuwien.ifs.feature.extraction.audio.data.FeatureVectorData;

/**
 * Copyright Vienna University of Technology
 * 
 * @author Thomas Lidy
 * @version $Id: SOMLibFormatWriter.java 179 2010-07-28 18:08:24Z mayer $ inspired by somtoolbox
 */
public class SOMLibFormatWriter extends AbstractDataWriter {

    private final String realFileName;

    private int numVectors = 0;

    private int matrixRows = 0;

    private int matrixCols = 0;

    /**
     * Constructor for SOMLibFormatWriter. if the outFileName ends with ".gz" gzip compression will
     * be enabled. If b
     * 
     * @param outFileName the output filename
     * @param gzipped set <code>true</code> to force gzip compression
     * @throws IOException
     */
    public SOMLibFormatWriter(String outFileName, boolean gzipped) throws IOException {
        super(outFileName + ".tmp");
        this.gzipped = gzipped;
        if (gzipped) {
            if (!outFileName.endsWith(".gz")) {
                outFileName = outFileName + ".gz";
            }
        }
        realFileName = outFileName;
    }

    private void writeHeader(BufferedWriter outFile) throws IOException {
        outFile.write("$TYPE vec");
        outFile.newLine();
        if (dataTypeID != null) {
            outFile.write("$DATA_TYPE " + dataTypeID);
            outFile.newLine();
        }
        if (matrixRows > 0 || matrixCols > 0) {
            outFile.write("$DATA_DIM " + matrixRows + "x" + matrixCols);
            outFile.newLine();
        }
        if (baseDirectory != null) {
            outFile.write("$DATA_DIR " + baseDirectory);
            outFile.newLine();
        }
        outFile.write("$XDIM " + numVectors);
        outFile.newLine();
        outFile.write("$YDIM 1");
        outFile.newLine();
        outFile.write("$VEC_DIM " + featureVectorLength);
        outFile.newLine();
    }

    @Override
    public void close() throws IOException {
        try {
            super.close();
        } catch (IOException e) {
            throw new IOException("Problem closing SOMLib vector file " + fileName + "! "
                    + e.getMessage());
        }

        try {
            BufferedWriter finalFile;
            if (gzipped) {
                finalFile = new BufferedWriter(new OutputStreamWriter(new GZIPOutputStream(
                        new FileOutputStream(realFileName))));
            } else {
                finalFile = new BufferedWriter(new FileWriter(realFileName));
            }

            writeHeader(finalFile);

            /* copy temp file to final File */
            String line;
            BufferedReader inFile = new BufferedReader(new FileReader(fileName));

            while ((line = inFile.readLine()) != null) {
                finalFile.write(line);
                finalFile.newLine();
            }
            inFile.close();
            finalFile.close();

            // boolean success = (new File(tempFileName)).delete();
            // System.out.println("del sucdess: " + success);
            new File(fileName).delete();
        } catch (IOException e) {
            throw new IOException("Problem writing temporary SOMLib vector file to final file "
                    + realFileName + "! " + e.getMessage());
        }
    }

    /*
     * public void writeVector(double[] featureVector, String fileLabel) throws
     * IOException { int vectorDim = featureVector.length;
     * checkVectorDim(vectorDim);
     * try { for (int ve=0; ve < vectorDim; ve++) { if (featureVector[ve]==0) {
     * outFile.write("0 "); } else {
     * outFile.write(formatDouble.format(featureVector[ve])+" "); } }
     * outFile.write(fileLabel); outFile.newLine(); numVectors++; } catch
     * (IOException e) { throw new
     * IOException("Problem writing SOMLib vector file " + tempFileName + "! " +
     * e.getMessage()); } }
     */

    // public void writeMatrix(double[][] matrix, String label) throws
    // IOException
    // {
    @Override
    public void writeVector(FeatureVectorData featureVector) throws IOException {
        double[][] matrix = featureVector.getMatrixData();
        int nRows = matrix.length;
        int nCols = matrix[0].length;
        int vectorDim = nRows * nCols;

        checkVectorDim(vectorDim);

        matrixRows = nRows;
        matrixCols = nCols;

        try {
            /* write matrix column-wise, as Matlab does when vectorizing */

            for (int c = 0; c < nCols; c++) {
                for (int r = 0; r < nRows; r++) {
                    outFile.write(formatDouble.format(matrix[r][c]) + " ");
                }
            }
            outFile.write(featureVector.getFileLabel());
            outFile.newLine();
            numVectors++;
        } catch (IOException e) {
            throw new IOException("Problem writing SOMLib vector file " + fileName + "! "
                    + e.getMessage());
        }
    }

}
