/*
 * Copyright 2004-2010 Information & Software Engineering Group (188/1)
 *                     Institute of Software Technology and Interactive Systems
 *                     Vienna University of Technology, Austria
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.ifs.tuwien.ac.at/mir/index.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package at.tuwien.ifs.feature.extraction.audio.util;

import java.util.ArrayList;

import at.tuwien.ifs.feature.extraction.audio.util.LogHandler.LogGrade;

/**
 * Static logging class. Was completely reworked when implementing multi-threaded extraction.
 * 
 * @author Jakob Frank
 * @author Thomas Lidy
 * @version 2.0
 */
public class MyLogger {

    private static ArrayList<LogHandler> handlers = new ArrayList<LogHandler>();

    /**
     * Whether the logging should be automatically flushed after each message. This is useful
     * specifically for single-threaded applications.
     */
    private static boolean autoFlush = false;

    /**
     * Initialises the logger for standard sequential logging, by registering a default
     * {@link PrintStreamLogger}, and by setting {@link #setAutoFlush(boolean)}.
     */
    public static void initForStandardLogging() {
        registerLogHandler(new PrintStreamLogger());
        setAutoFlush(true);
    }

    /**
     * Register a log-handler.
     * 
     * @param handler The handler.
     * @return <code>true</code> on success.
     */
    public static boolean registerLogHandler(LogHandler handler) {
        return handlers.add(handler);
    }

    public static void setAutoFlush(boolean autoFlush) {
        MyLogger.autoFlush = autoFlush;
    }

    /**
     * Remove a log-handler. The log-handler will be flushed before removing.
     * 
     * @param handler The log-handler to remove
     * @return <code>true</code> on success.
     */
    public static boolean removeLogHandler(LogHandler handler) {
        boolean res = handlers.remove(handler);
        if (res) {
            handler.flushLog();
        }
        return res;
    }

    private static void sendLog(String message, LogGrade grade) {
        for (LogHandler handler : handlers) {
            handler.appendLog(message, grade);
            if (autoFlush) {
                // TODO: maybe call flushAll()
                // or is automatically flushing always everything equivalent?
                flush();
            }
        }
    }

    /**
     * Log an error.
     * 
     * @param error The error-string.
     */
    public static void error(String error) {
        sendLog(error, LogGrade.ERROR);
    }

    /**
     * Log some extra info.
     * 
     * @param extra The extra-info.
     */
    public static void extrainfo(String extra) {
        sendLog(extra, LogGrade.EXTRA_INFO);
    }

    /**
     * Log some extra info to the previous line.
     * 
     * @param extra the extra-info.
     */
    public static void extrainfoNoNewLine(String extra) {
        sendLog(extra, LogGrade.SAME);
    }

    /**
     * Log an info line.
     * 
     * @param info the info.
     */
    public static void info(String info) {
        sendLog(info, LogGrade.INFO);
    }

    /**
     * Log a severe error.
     * 
     * @param severe the error.
     */
    public static void severe(String severe) {
        sendLog(severe, LogGrade.SEVERE);
    }

    /**
     * Log a warning.
     * 
     * @param warning the warning.
     */
    public static void warning(String warning) {
        sendLog(warning, LogGrade.WARNING);
    }

    /**
     * Flush all logging handlers.
     */
    public static void flush() {
        for (LogHandler handler : handlers) {
            handler.flushLog();
        }
    }

    public static void flushAll() {
        for (LogHandler handler : handlers) {
            handler.flushAllLogs();
        }
    }

    /**
     * Set the log-grade for all registered handlers.
     * 
     * @param grade the new grade.
     */
    public static void setGlobalLogGrade(LogGrade grade) {
        for (LogHandler handler : handlers) {
            handler.setLogGrade(grade);
        }
    }
}
